#ifndef USERLIST_H
#define USERLIST_H

#include <qobject.h>
#include <qvaluelist.h>
#include <qtimer.h>
#include <qhostaddress.h>
#include <qdns.h>
#include <qptrlist.h>

#include "gadu.h"

class UserList;
class UserListElement;

class DnsHandler : public QObject
{
	Q_OBJECT

	UserListElement &Ule;

	public:
		DnsHandler(UserListElement &ule);
		~DnsHandler();

		static int counter;

	private:
		QDns DnsResolver;

	private slots:
		void resultsReady();
};

class UserListElement
{
	private:
		QString Group;
		QString FirstName;
		QString LastName;
		QString NickName;
		QString AltNick;
		QString Mobile;
		QString Email;
		UinType Uin;
		UserStatus *Stat;
		int MaxImageSize;
		bool Anonymous;
		QHostAddress Ip;
		QString DnsName;
		short Port;
		int Version;
		bool Blocking;
		bool OfflineTo;
		bool Notify;

		UserList *Parent;
		friend class UserList;

	public:
		UserListElement(UserList* parent);
		UserListElement(const UserListElement &copyMe);
		UserListElement();
		virtual ~UserListElement();

		void refreshDnsName();

		void operator = (const UserListElement &copyMe);

		QString group() const;
		void setGroup(const QString& group);

		QString firstName() const;
		void setFirstName(const QString &firstName);

		QString lastName() const;
		void setLastName(const QString &lastName);

		QString nickName() const;
		void setNickName(const QString &nickName);

		QString altNick() const;
		void setAltNick(const QString &altNick);

		QString mobile() const;
		void setMobile(const QString &mobile);

		QString email() const;
		void setEmail(const QString &email);

		UinType uin() const;
		void setUin(const UinType &uin);

		const UserStatus & status() const;
		UserStatus & status();

		int maxImageSize() const;
		void setMaxImageSize(const int maxImageSize);

		bool isAnonymous() const;
		void setAnonymous(const bool anonymous);

		const QHostAddress & ip() const;
		QHostAddress & ip();

		QString dnsName() const;
		void setDnsName(const QString &dnsName);

		short port() const;
		void setPort(short port);

		int version() const;
		void setVersion(const int version);

		bool blocking() const;
		void setBlocking(const bool blocking);

		bool offlineTo() const;
		void setOfflineTo(const bool offlineTo);

		bool notify() const;
		void setNotify(const bool notify);
};

/**
	Klasa reprezentujca list kontaktw.
	indexami s altnicki konwertowane tak,
	aby wszystkie litery byy mae.
	Dziki temu nie mog istnie dwa kontakty
	o tych samych altnickach nawet, jeli
	rni si wielkociami liter.
	Pole Altnick w UserListElement zawiera natomiast
	oryginalny altnick.
	
	UWAGA! Uywanie metod klasy QMap nie jest zalecane,
	trzeba wtedy pamita, e jako klucza uywamy altnickw
	odpowiednio skonwertowanych, e zmiany powinny generowa
	odpowiednie sygnay, itd. Duo lepiej uywa specjalnie
	przygotowanych metod klasy UserList.
	
	TODO: Zgodnie z tym co powyej z QMap powinnimy dziedziczy
	prywatnie albo niech to bdzie pole wewntrz klasy. W tej
	chwili nie chc ama api freeze w 0.4.0.
**/
class UserList : public QObject, public QMap<QString,UserListElement>
{
	Q_OBJECT

	protected:
		friend class UserListElement;

	public:
		UserList();
		UserList(const UserList& source);
		~UserList();
		UserList& operator=(const UserList& userlist);

		UserListElement& byUin(UinType uin);
		UserListElement& byNick(const QString& nickname);
		/**
			Wyszukuje kontakt po altnicku. Mae i
			due litery nie maj znaczenia.
		**/
		UserListElement& byAltNick(const QString& altnick);
		/**
			Wyszukuje kontakt po numerze uin.
			Jezeli nie znajdzie, zwracany jest
			UserListElement tylko z uin i altnick == uin
		**/
		UserListElement byUinValue(UinType uin);
		/**
			Sprawdza czy istnieje kontakt o danym numerze uin.
		**/
		bool containsUin(UinType uin) const;
		/**
			Sprawdza czy istnieje kontakt o danym altnicku.
			Mae i due litery nie maj znaczenia.
		**/
		bool containsAltNick(const QString& altnick) const;

	public slots:
		/**
			Dodaje do listy kontakt o podanych danych.
			Generuje nastpnie sygnay userDataChanged()
			i modified().
		**/
		void addUser(const UserListElement& ule);
		/**
			Dodaje do listy "anonimowy" kontakt o podanym
			numerze uin. Numer ten wykorzystywany jest jako
			altnick. Wywouje metod addUser(), wic
			generowane s sygnay userDataChanged() i modified().
		**/
		void addAnonymous(UinType uin);
		/**
			Usuwa kontakt o podanym altnicku. Mae i
			due litery nie maj znaczenia.
			Tu przed usuniciem kontaktu z listy generuje
			sygna userDataChanged(). Po usuniciu generowany
			jest sygna modified().
		**/
		void removeUser(const QString& altnick);
		/**
			Zmienia dane kontaktu o podanym altnicku. Mae i
			due litery w zmiennej old_altnick nie maj znaczenia.
			Po wykonaniu zmian wymusza ich uwzgldnienie przez
			wszystkie obiekty klasy UserBox wywoujc metod
			UserBox::all_renameUser() i UserBox::all_refresh().
			Na koniec generuje sygnay userDataChanged() i modified().
		**/		
		void changeUserInfo(const QString& old_altnick, const UserListElement& new_data);
		/**
			Zapisuje list kontaktw do pliku o podanej nazwie
			(wraz ze ciek). Domylnie jest to plik "userlist"
			w katalogu z ustawieniami.
			Utrzymywana jest kompatybilno z programem EKG,
			z tej przyczyny dodatkowe informacje jak to czy
			blokujemy uytkownika, czy chcemy by powiadamiani
			o zmianach jego statusu itp. s zapisywane do
			oddzielnego pliku "userattribs" w katalogu z
			ustawieniami.
			Kontakty "anonimowe" s pomijane.
			
			TODO: To gdzie zapisujemy dodatkowe atrybutami te
			powinno by chyba podawane jako argument? Inaczej
			jeli chcemy zapisa list gdzie indziej to nadpisz
			nam si atrybuty :/
		**/
		bool writeToFile(QString filename = "");
		/**
			Wczytuje list kontaktw z plikw "userlist"
			i "userattribs" z katalogu z ustawieniami.
			Metoda ta jest przeznaczona do przywracania
			danych zapisanych za pomoc writeToFile().
		**/		
		bool readFromFile();
		void merge(const UserList &userlist);

	signals:
		/**
			Sygna generowany po zakoczeniu zmian w licie
			kontaktw.
		**/
		void modified();
		void dnsNameReady(UinType);
		/**
			Sygna jest generowany jeli cokolwiek zmienio
			si w danych ktrego z kontaktw na licie
			(wtedy przekazywane s argumenty oldData i newData).
			Zmiana moe nastpi przez bezporednie uycie ktrej
			z metod klasy UserListElement.
			Sygna przekazuje rwnie informacj o tym, e nowy
			kontakt pojawi si na licie (wtedy oldData == NULL)
			lub e zosta z listy usunity (wtedy newData = NULL).
		**/
		void userDataChanged(const UserListElement* const oldData, const UserListElement* const newData);
};

extern UserList userlist;

#endif

